/*
 * Copyright 2010 pilot <pilot@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>

#include <ruby.h>
#include "arirang.h"


extern int sslflags;
extern short int CONNECT_TIMEOUT;
extern int scan_count;

extern int ruby;
extern char *ariruby;
extern FILE *rubyfd;

extern int proxy;
extern char *proxyhost;
extern int proxyport;
char rscanhost[128];
int rport;

int sock;
int nb = 0;
struct sockaddr_in http;
struct hostent *he;
char *osfinger = NULL;
char *headbuf = NULL;
int recv_size = 2048;
int recv_flag = 0;

SSL_CTX *ctx = NULL;
SSL *ssl = NULL;
int err;
int sread = 0;


void
connout_r(int sig)
{
        exit(0);
}


int
recvtimeout_r(int s, char *buf, int len, int timeout)
{
    fd_set  fds;
    int n;
    struct timeval  tv;

    FD_ZERO(&fds);
    FD_SET(s, &fds);
    tv.tv_sec = timeout;
    tv.tv_usec = 0;

    n = select(s + 1, &fds, NULL, NULL, &tv);
    
    if (n == 0)
        return -2;
    if (n == -1)
        return -1;

#ifdef DEBUG
    printf("recv_flag = %d \n", recv_flag);
#endif

    return recv(s, buf, len, recv_flag);
}


void
ariruby_init(char *rubyfile)
{
    ruby_init();
    ruby_init_loadpath();
    rb_load_file(rubyfile);
}

void
ariruby_run(void)
{
    ruby_exec();
}

void
ariruby_exit(void)
{
    ruby_finalize();
}


static VALUE 
ariconnect(VALUE self) 
{

    if (proxy) {
        sock = proxy_connect(rscanhost, rport, proxyhost, proxyport, 1);
    }


    if (!proxy) {
        if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
            perror("socket");
            close(sock);
            exit(0);
        }

        signal(SIGALRM, connout_r);
        alarm(2);
        if (connect(sock, (struct sockaddr *) & http, sizeof(struct sockaddr)) == -1) {
            close(sock);
            exit(0);
        }
        alarm(0); 
    }

    if (sslflags) {
        ctx = ssl_init_ctx();
        ssl = ssl_init_ssl(ctx, sock);
    }
#ifdef DEBUG
    printf("connect \n");
#endif
    return self;
}

static VALUE 
arisend(VALUE self, VALUE data) 
{
    char *rbuf;
    rbuf = StringValuePtr(data);

    if ((osfinger = (char *)malloc(strlen(rbuf) + 1)) == NULL)  {
        perror("malloc failed");
        close(sock);
        exit(0);
    }
    strcpy(osfinger, rbuf);

    if (strlen(rbuf) == 0) {
        fprintf(stderr, "please set a data in arisend function \n");
        exit(0);
    }
#ifdef DEBUG
    printf("send data  = %s\n", osfinger);
    printf("recv data size = %d \n", recv_size);
#endif

    if ((headbuf = (char *)malloc(recv_size + 1)) == NULL)  {
        perror("malloc failed");
        close(sock);
        exit(0);
    }

    bzero(headbuf, recv_size);

    if (sslflags) {
        err = SSL_write(ssl, osfinger, strlen(osfinger));
        if (err <= 0) {
        close(sock);
/*        fprintf(stdout, "%s - ssl write time out \n", scanhost); */
        exit(0);
        }
#ifdef DEBUG
     printf("error = %d\n", err);
#endif
     sread = ssl_read_timeout(sock, ssl, headbuf, recv_size, CONNECT_TIMEOUT);

        if (sread <= 0 ) {
            close(sock);
/*            fprintf(stdout, "%s - ssl read time out \n", scanhost); */
        exit(0);
        }
#ifdef DEBUG
     printf("sread  = %d \n", sread);
#endif
    }
    else {
    send(sock, osfinger, strlen(osfinger), 0);
    nb = recvtimeout_r(sock, headbuf, recv_size, CONNECT_TIMEOUT);
    }

    return (rb_str_new2(headbuf));

}


void
rubyscan(char *scanhost, unsigned int port, short int option)
{

    ID ari_send_ID;
    ID arirang_ID;

    VALUE rb_cArirang;
    VALUE rb_arirang;
    VALUE ari_port;
    VALUE ari_recv;
    VALUE ari_recvflag;
    VALUE rhost;
   
    if (!proxy) {
        if ((he = gethostbyname(scanhost)) == NULL) {
            perror("host not found");
            /* close(sock);  */
            exit(0);
        }
    }
    signal(SIGINT, connout_r);

    /*ariruby_init(ariruby); */

    arirang_ID = rb_intern("Arirang");
    ari_send_ID = rb_intern( "asadal" );

    /*ariruby_run();*/

    rb_cArirang = rb_const_get(rb_cObject, arirang_ID);
    rb_arirang = rb_class_new_instance(0, NULL, rb_cArirang);
   
    rb_define_method(rb_cArirang, "ariconnect", ariconnect, 0);
    rb_define_method(rb_cArirang, "arisend", arisend, 1);

    ari_port = rb_ivar_get(rb_cArirang, rb_intern("@ari_port"));
    ari_recv = rb_ivar_get(rb_cArirang, rb_intern("@ari_recvsize"));
    ari_recvflag = rb_ivar_get(rb_cArirang, rb_intern("@ari_recvflag"));

/*
    ari_hash = rb_hash_aref(ari_info, rb_str_new2("description"));
    ari_array = rb_ary_entry(ari_info, 0);
*/

    if (ari_port != Qnil)
        port = NUM2INT(ari_port);

    if (proxy) {
        snprintf(rscanhost, sizeof(rscanhost), "%s", scanhost);
        rport = port;
        #ifdef DEBUG
        printf("rscanhost = %s, rport = %d \n", rscanhost, rport);
        #endif
    }

    if (ari_recv != Qnil)  
        recv_size = NUM2INT(ari_recv);

    if (ari_recvflag != Qnil)
        recv_flag = NUM2INT(ari_recvflag);

    switch (recv_flag) {
        case 1:
            recv_flag = MSG_OOB;
            break;
        case 2:
            recv_flag = MSG_PEEK;
            break;
        case 3:
            recv_flag = MSG_WAITALL;
            break;
        case 4:
            recv_flag = MSG_DONTWAIT;
            break;
        default:
            recv_flag = 0;
    }

    if (!proxy) {
        http.sin_family = AF_INET;
        http.sin_port = htons(port);
        http.sin_addr = *((struct in_addr *) he->h_addr);
        bzero(&(http.sin_zero), 8);
    }

    rhost = rb_str_new2(scanhost);
    rb_define_variable("$ari_host", &rhost);

    rb_funcall(rb_arirang, ari_send_ID, 0 );

    if (sslflags) {
        if (err <= 0) {
        close(sock);
        fprintf(stdout, "%s - ssl write time out \n", scanhost);
        exit(0);
        }

    ssl_error(ssl, err, scanhost);
#ifdef DEBUG
    printf("error = %d\n", err);
#endif
    }

    if (sslflags) {
        if (sread <= 0 ) {
            fprintf(stdout, "%s - ssl read time out \n", scanhost);
            close(sock);
            exit(0);
        }
    }
    
    
    if (osfinger != NULL)
        free(osfinger);

    if (headbuf != NULL)
        free(headbuf);

    if (nb == -1 || nb == -2 ) {
        fprintf(stdout, "%s - maybe unknown network device or filter by firewall \n", scanhost);
        close(sock);
        exit(0);
    }


    if (sslflags)
        ssl_free(ctx, ssl);

    ariruby_exit();

    close(sock);

}



void
rubyscanW(unsigned long host, unsigned int port, unsigned short int type, short int option)
{

    ID ari_send_ID;
    ID arirang_ID;

    VALUE rb_cArirang;
    VALUE rb_arirang;
    VALUE ari_port;
    VALUE ari_recv;
    VALUE ari_recvflag;
    VALUE rhost;

    signal(SIGINT, connout_r);

    /*ariruby_init(ariruby);*/

    arirang_ID = rb_intern("Arirang");
    ari_send_ID = rb_intern( "asadal" );

    /*ariruby_run();*/

    rb_cArirang = rb_const_get(rb_cObject, arirang_ID);
    rb_arirang = rb_class_new_instance(0, NULL, rb_cArirang);

    rb_define_method(rb_cArirang, "ariconnect", ariconnect, 0);
    rb_define_method(rb_cArirang, "arisend", arisend, 1);

    ari_port = rb_ivar_get(rb_cArirang, rb_intern("@ari_port"));
    ari_recv = rb_ivar_get(rb_cArirang, rb_intern("@ari_recvsize"));
    ari_recvflag = rb_ivar_get(rb_cArirang, rb_intern("@ari_recvflag"));
/*
    ari_hash = rb_hash_aref(ari_info, rb_str_new2("description"));
    ari_array = rb_ary_entry(ari_info, 0);
*/
    
    if (ari_port != Qnil)
        port = NUM2INT(ari_port);

    if (ari_recv != Qnil)  
        recv_size = NUM2INT(ari_recv);

    if (ari_recvflag != Qnil)
        recv_flag = NUM2INT(ari_recvflag);

    switch (recv_flag) {
        case 1:
            recv_flag = MSG_OOB;
            break;
        case 2:
            recv_flag = MSG_PEEK;
            break;
        case 3:
            recv_flag = MSG_WAITALL;
            break;
        case 4:
            recv_flag = MSG_DONTWAIT;
            break;
        default:
            recv_flag = 0;
    }

    http.sin_family = AF_INET;
    http.sin_port = htons(port);
    http.sin_addr.s_addr = htonl(host);


    if (proxy) {
        snprintf(rscanhost, sizeof(rscanhost), "%s", inet_ntoa(http.sin_addr));
        rport = port;
#ifdef DEBUG
    printf("rscanhost = %s, rport = %d \n", rscanhost, rport);
#endif
    }


#ifdef DEBUG
    printf("Debug %s\n", inet_ntoa(http.sin_addr)); 
#endif

    rhost = rb_str_new2(inet_ntoa(http.sin_addr));
    rb_define_variable("$ari_host", &rhost);

    rb_funcall(rb_arirang, ari_send_ID, 0);


    if (sslflags) {
        if (err <= 0) {
        close(sock);
        fprintf(stdout, "%s - ssl write time out \n", inet_ntoa(http.sin_addr));
        exit(0);
        }

    ssl_error(ssl, err, inet_ntoa(http.sin_addr)); 
#ifdef DEBUG
    printf("error = %d\n", err); 
#endif
    }

    if (sslflags) {
        if (sread <= 0 ) {
            fprintf(stdout, "%s - ssl read time out \n", inet_ntoa(http.sin_addr));
            close(sock);
            exit(0);
        }
    }
#ifdef DEBUG
    printf("sread  = %d, buf size= %d \n", sread, strlen(headbuf)); 
#endif

    if (osfinger != NULL)
        free(osfinger);

    if (headbuf != NULL)
        free(headbuf);

    if (nb == -1 || nb == -2) { 
#if !(defined(__OpenBSD__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__APPLE__))
    alarm(0);
#endif
    close(sock);
    exit(0);
    }

    if (strstr(headbuf, "Server Error") != NULL) {
#if !(defined(__OpenBSD__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__APPLE__))
    alarm(0);
#endif
    close(sock);
    exit(0);
    }


    if (sslflags)
        ssl_free(ctx, ssl);

    ariruby_exit();

    close(sock);
}
